from __future__ import annotations

"""Utility functions for geometry, acoustics, and tensor handling."""

import math
import warnings
from dataclasses import dataclass
from typing import Iterable, Optional, Tuple

import torch
from torch import Tensor


_DEF_SPEED_OF_SOUND = 343.0


def as_tensor(
    value: Tensor | Iterable[float] | float | int,
    *,
    device: Optional[torch.device | str] = None,
    dtype: Optional[torch.dtype] = None,
) -> Tensor:
    """Convert a value to a tensor while preserving device/dtype when possible."""
    if isinstance(device, str):
        device = resolve_device(device)
    if torch.is_tensor(value):
        out = value
        if device is not None:
            out = out.to(device)
        if dtype is not None:
            out = out.to(dtype)
        return out
    return torch.as_tensor(value, device=device, dtype=dtype)


def resolve_device(
    device: Optional[torch.device | str],
    *,
    prefer: Tuple[str, ...] = ("cuda", "mps", "cpu"),
) -> torch.device:
    """Resolve a device string (including 'auto') into a torch.device.

    Falls back to CPU when the requested backend is unavailable.
    """
    if device is None:
        return torch.device("cpu")
    if isinstance(device, torch.device):
        return device

    dev = str(device).lower()
    if dev == "auto":
        for backend in prefer:
            if backend == "cuda" and torch.cuda.is_available():
                return torch.device("cuda")
            if backend == "mps" and torch.backends.mps.is_available():
                return torch.device("mps")
            if backend == "cpu":
                return torch.device("cpu")
        return torch.device("cpu")

    if dev.startswith("cuda"):
        if torch.cuda.is_available():
            return torch.device(device)
        warnings.warn("CUDA not available; falling back to CPU.", RuntimeWarning)
        return torch.device("cpu")
    if dev == "mps":
        if torch.backends.mps.is_available():
            return torch.device("mps")
        warnings.warn("MPS not available; falling back to CPU.", RuntimeWarning)
        return torch.device("cpu")
    if dev == "cpu":
        return torch.device("cpu")

    return torch.device(device)


@dataclass(frozen=True)
class DeviceSpec:
    """Resolve device + dtype defaults consistently."""

    device: Optional[torch.device | str] = None
    dtype: Optional[torch.dtype] = None
    prefer: Tuple[str, ...] = ("cuda", "mps", "cpu")

    def resolve(self, *values) -> Tuple[torch.device, torch.dtype]:
        """Resolve device/dtype from inputs with overrides."""
        tensor_device: Optional[torch.device] = None
        tensor_dtype: Optional[torch.dtype] = None
        for value in values:
            if torch.is_tensor(value):
                if tensor_device is None:
                    tensor_device = value.device
                if tensor_dtype is None:
                    tensor_dtype = value.dtype

        if isinstance(self.device, str) and self.device.lower() == "auto":
            device = tensor_device or resolve_device("auto", prefer=self.prefer)
        elif self.device is None:
            device = tensor_device or torch.device("cpu")
        else:
            device = resolve_device(self.device, prefer=self.prefer)

        if self.dtype is None:
            dtype = tensor_dtype or torch.float32
        else:
            dtype = self.dtype
        return device, dtype


def infer_device_dtype(
    *values,
    device: Optional[torch.device | str] = None,
    dtype: Optional[torch.dtype] = None,
) -> Tuple[torch.device, torch.dtype]:
    """Infer device/dtype from inputs with optional overrides."""
    return DeviceSpec(device=device, dtype=dtype).resolve(*values)


def ensure_dim(size: Tensor) -> Tensor:
    """Validate room size dimensionality (2D or 3D)."""
    if size.ndim != 1 or size.numel() not in (2, 3):
        raise ValueError("room size must be a 1D tensor of length 2 or 3")
    return size


def extend_size(size: Tensor, dim: int) -> Tensor:
    """Extend 2D room size to 3D by adding a dummy z dimension."""
    if size.numel() == dim:
        return size
    if size.numel() == 2 and dim == 3:
        pad = torch.tensor([1.0], device=size.device, dtype=size.dtype)
        return torch.cat([size, pad])
    raise ValueError("unsupported room dimension")


def estimate_beta_from_t60(
    size: Tensor,
    t60: float,
    *,
    device: Optional[torch.device | str] = None,
    dtype: Optional[torch.dtype] = None,
) -> Tensor:
    """Estimate reflection coefficients from T60 using Sabine's formula."""
    if t60 <= 0:
        raise ValueError("t60 must be positive")
    size = as_tensor(size, device=device, dtype=dtype)
    size = ensure_dim(size)
    dim = size.numel()
    if dim == 2:
        lx, ly = size.tolist()
        lz = 1.0
        volume = lx * ly * lz
        surface = 2.0 * (lx + ly) * lz
        alpha = 0.161 * volume / (t60 * surface)
        alpha = max(0.0, min(alpha, 0.999))
        beta = math.sqrt(1.0 - alpha)
        return torch.full((4,), beta, device=size.device, dtype=size.dtype)
    size = extend_size(size, 3)
    lx, ly, lz = size.tolist()
    volume = lx * ly * lz
    surface = 2.0 * (lx * ly + ly * lz + lx * lz)
    alpha = 0.161 * volume / (t60 * surface)
    alpha = max(0.0, min(alpha, 0.999))
    beta = math.sqrt(1.0 - alpha)
    return torch.full((6,), beta, device=size.device, dtype=size.dtype)


def estimate_t60_from_beta(
    size: Tensor,
    beta: Tensor,
    *,
    device: Optional[torch.device | str] = None,
    dtype: Optional[torch.dtype] = None,
) -> float:
    """Estimate T60 from reflection coefficients using Sabine's formula."""
    size = as_tensor(size, device=device, dtype=dtype)
    size = ensure_dim(size)
    beta = as_tensor(beta, device=size.device, dtype=size.dtype)
    dim = size.numel()
    if dim == 2:
        if beta.numel() != 4:
            raise ValueError("beta must have 4 elements for 2D t60 estimation")
        lx, ly = size.tolist()
        lz = 1.0
        volume = lx * ly * lz
        surfaces = torch.tensor(
            [ly * lz, ly * lz, lx * lz, lx * lz],
            device=size.device,
            dtype=size.dtype,
        )
        alpha = 1.0 - beta**2
        absorption = torch.sum(surfaces * alpha).item()
        if absorption <= 0.0:
            return float("inf")
        return 0.161 * volume / absorption
    size = extend_size(size, 3)
    if beta.numel() != 6:
        raise ValueError("beta must have 6 elements for t60 estimation")
    lx, ly, lz = size.tolist()
    volume = lx * ly * lz
    surfaces = torch.tensor(
        [ly * lz, ly * lz, lx * lz, lx * lz, lx * ly, lx * ly],
        device=size.device,
        dtype=size.dtype,
    )
    alpha = 1.0 - beta**2
    absorption = torch.sum(surfaces * alpha).item()
    if absorption <= 0.0:
        return float("inf")
    return 0.161 * volume / absorption


def orientation_to_unit(orientation: Tensor, dim: int) -> Tensor:
    """Convert orientation representation to unit vectors in 2D/3D."""
    if dim == 2:
        if orientation.ndim == 0:
            angle = orientation
            vec = torch.stack([torch.cos(angle), torch.sin(angle)])
            return normalize_orientation(vec)
        if orientation.shape[-1] == 1:
            angle = orientation.squeeze(-1)
            vec = torch.stack([torch.cos(angle), torch.sin(angle)], dim=-1)
            return normalize_orientation(vec)
        if orientation.ndim == 1 and orientation.numel() != 2:
            angle = orientation
            vec = torch.stack([torch.cos(angle), torch.sin(angle)], dim=-1)
            return normalize_orientation(vec)
        if orientation.shape[-1] == 2:
            return normalize_orientation(orientation)
        raise ValueError("2D orientation must be angle or 2D vector")
    if dim == 3:
        if orientation.shape[-1] == 3:
            return normalize_orientation(orientation)
        if orientation.shape[-1] == 2:
            azimuth = orientation[..., 0]
            elevation = orientation[..., 1]
            x = torch.cos(elevation) * torch.cos(azimuth)
            y = torch.cos(elevation) * torch.sin(azimuth)
            z = torch.sin(elevation)
            vec = torch.stack([x, y, z], dim=-1)
            return normalize_orientation(vec)
        raise ValueError("3D orientation must be vector or (azimuth, elevation)")
    raise ValueError("unsupported dimension for orientation")


def att2t_sabine_estimation(att_db: float, t60: float) -> float:
    """Convert attenuation (dB) to time based on T60."""
    if t60 <= 0:
        raise ValueError("t60 must be positive")
    if att_db <= 0:
        raise ValueError("att_db must be positive")
    return (att_db / 60.0) * t60


def att2t_SabineEstimation(att_db: float, t60: float) -> float:
    """Legacy alias for att2t_sabine_estimation."""
    return att2t_sabine_estimation(att_db, t60)


def beta_SabineEstimation(room_size: Tensor, t60: float) -> Tensor:
    """Legacy alias for estimate_beta_from_t60."""
    return estimate_beta_from_t60(room_size, t60)


def t2n(tmax: float, room_size: Tensor, c: float = _DEF_SPEED_OF_SOUND) -> Tensor:
    """Estimate image counts per dimension needed to cover tmax."""
    if tmax <= 0:
        raise ValueError("tmax must be positive")
    size = as_tensor(room_size)
    size = ensure_dim(size)
    # number of images in each dimension needed to cover the maximum distance
    # uses the same heuristic as gpuRIR: n = ceil(tmax * c / room_size)
    n = torch.ceil((tmax * c) / size).to(torch.int64)
    return n


def normalize_orientation(orientation: Tensor, *, eps: float = 1e-8) -> Tensor:
    """Normalize orientation vectors with numerical stability."""
    norm = torch.linalg.norm(orientation, dim=-1, keepdim=True)
    norm = torch.clamp(norm, min=eps)
    return orientation / norm



import math

import numpy as np
import pytest
import torch
import torch.nn.functional as F

from torchrir import MicrophoneArray, Room, Source, fft_convolve, simulate_rir

pra = pytest.importorskip("pyroomacoustics")


def _align_by_xcorr(a: torch.Tensor, b: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    a = a.detach().cpu()
    b = b.detach().cpu()
    corr = F.conv1d(a.view(1, 1, -1), b.view(1, 1, -1), padding=b.numel() - 1)
    corr = corr.view(-1)
    idx = torch.argmax(torch.abs(corr)).item()
    lag = idx - (b.numel() - 1)
    if lag >= 0:
        a = a[lag:]
    else:
        b = b[-lag:]
    min_len = min(a.numel(), b.numel())
    return a[:min_len], b[:min_len]


def _rel_l2(a: torch.Tensor, b: torch.Tensor) -> float:
    return torch.linalg.norm(a - b).item() / (torch.linalg.norm(b).item() + 1e-8)


def test_rir_and_convolved_signal_close():
    fs = 16000
    room_dim = [6.0, 4.0, 3.0]
    src = [1.0, 1.5, 1.2]
    mic = [3.0, 2.0, 1.2]
    max_order = 3
    beta = 0.9

    # pyroomacoustics RIR
    pra.constants.set("rir_hpf_enable", False)
    absorption = 1.0 - beta**2
    material = pra.Material(absorption) if hasattr(pra, "Material") else None
    pra_room = pra.ShoeBox(room_dim, fs=fs, max_order=max_order, materials=material)
    pra_room.add_source(src)
    mic_locs = torch.tensor(mic).view(3, 1).numpy()
    mic_array = pra.MicrophoneArray(mic_locs, fs)
    pra_room.add_microphone_array(mic_array)
    pra_room.compute_rir()
    pra_rir_np = np.array(pra_room.rir[0][0], dtype=np.float32, copy=True)
    pra_rir = torch.from_numpy(pra_rir_np)

    # torchrir RIR
    room = Room.shoebox(size=room_dim, fs=fs, beta=[beta] * 6)
    sources = Source.positions([src])
    mics = MicrophoneArray.positions([mic])
    torch_rir = simulate_rir(
        room=room,
        sources=sources,
        mics=mics,
        max_order=max_order,
        nsample=pra_rir.numel(),
        directivity="omni",
    )[0, 0]

    # Align and compare RIRs
    a, b = _align_by_xcorr(torch_rir, pra_rir)
    rir_err = _rel_l2(a, b)
    assert rir_err < 0.5

    # Convolve a test signal and compare
    t = torch.linspace(0.0, 1.0, fs, dtype=torch.float32)
    test_sig = torch.sin(2.0 * math.pi * 440.0 * t)

    y_torch = fft_convolve(test_sig, a)
    y_pra = fft_convolve(test_sig, b)

    y_a, y_b = _align_by_xcorr(y_torch, y_pra)
    sig_err = _rel_l2(y_a, y_b)
    assert sig_err < 0.5



import pytest
import torch

from torchrir import ISMSimulator, MicrophoneArray, RIRResult, Room, Scene, SimulationConfig, Source


def test_scene_validate_static():
    room = Room.shoebox(size=[5.0, 4.0, 3.0], fs=16000, beta=[0.9] * 6)
    sources = Source.positions([[1.0, 1.0, 1.0]])
    mics = MicrophoneArray.positions([[2.0, 1.5, 1.0]])
    scene = Scene(room=room, sources=sources, mics=mics)
    scene.validate()


def test_scene_validate_dynamic_shapes():
    room = Room.shoebox(size=[5.0, 4.0, 3.0], fs=16000, beta=[0.9] * 6)
    sources = Source.positions([[1.0, 1.0, 1.0]])
    mics = MicrophoneArray.positions([[2.0, 1.5, 1.0]])
    src_traj = torch.tensor(
        [
            [[1.0, 1.0, 1.0]],
            [[1.5, 1.0, 1.0]],
        ]
    )
    mic_traj = torch.tensor(
        [
            [[2.0, 1.5, 1.0]],
            [[2.2, 1.5, 1.0]],
        ]
    )
    scene = Scene(room=room, sources=sources, mics=mics, src_traj=src_traj, mic_traj=mic_traj)
    scene.validate()


def test_scene_validate_mismatch_time():
    room = Room.shoebox(size=[5.0, 4.0, 3.0], fs=16000, beta=[0.9] * 6)
    sources = Source.positions([[1.0, 1.0, 1.0]])
    mics = MicrophoneArray.positions([[2.0, 1.5, 1.0]])
    src_traj = torch.tensor(
        [
            [[1.0, 1.0, 1.0]],
            [[1.5, 1.0, 1.0]],
        ]
    )
    mic_traj = torch.tensor(
        [
            [[2.0, 1.5, 1.0]],
            [[2.2, 1.5, 1.0]],
            [[2.4, 1.5, 1.0]],
        ]
    )
    scene = Scene(room=room, sources=sources, mics=mics, src_traj=src_traj, mic_traj=mic_traj)
    with pytest.raises(ValueError):
        scene.validate()


def test_rir_result_container():
    room = Room.shoebox(size=[5.0, 4.0, 3.0], fs=16000, beta=[0.9] * 6)
    sources = Source.positions([[1.0, 1.0, 1.0]])
    mics = MicrophoneArray.positions([[2.0, 1.5, 1.0]])
    scene = Scene(room=room, sources=sources, mics=mics)
    config = SimulationConfig(max_order=1, tmax=0.1)
    rirs = torch.zeros((1, 1, 256))
    result = RIRResult(rirs=rirs, scene=scene, config=config, seed=123)
    assert result.rirs.shape == (1, 1, 256)


def test_ism_simulator_static():
    room = Room.shoebox(size=[4.0, 3.0, 2.5], fs=16000, beta=[0.9] * 6)
    sources = Source.positions([[1.0, 1.0, 1.0]])
    mics = MicrophoneArray.positions([[2.0, 1.5, 1.0]])
    scene = Scene(room=room, sources=sources, mics=mics)
    config = SimulationConfig(max_order=1, tmax=0.05)
    result = ISMSimulator().simulate(scene, config)
    assert result.rirs.ndim == 3



from __future__ import annotations

"""Matplotlib-based plotting helpers for room scenes."""

from typing import Any, Iterable, Optional, Sequence, Tuple

import torch
from torch import Tensor

from .room import MicrophoneArray, Room, Source
from .utils import as_tensor, ensure_dim


def plot_scene_static(
    *,
    room: Room | Sequence[float] | Tensor,
    sources: Source | Tensor | Sequence,
    mics: MicrophoneArray | Tensor | Sequence,
    ax: Any | None = None,
    title: Optional[str] = None,
    show: bool = False,
):
    """Plot a static room with source and mic positions."""
    plt, ax = _setup_axes(ax, room)

    size = _room_size(room, ax)
    _draw_room(ax, size)

    src = _extract_positions(sources, ax)
    mic = _extract_positions(mics, ax)

    _scatter_positions(ax, src, label="sources", marker="^")
    _scatter_positions(ax, mic, label="mics", marker="o")

    if title:
        ax.set_title(title)
    ax.legend(loc="best")
    if show:
        plt.show()
    return ax


def plot_scene_dynamic(
    *,
    room: Room | Sequence[float] | Tensor,
    src_traj: Tensor | Sequence,
    mic_traj: Tensor | Sequence,
    step: int = 1,
    ax: Any | None = None,
    title: Optional[str] = None,
    show: bool = False,
):
    """Plot source and mic trajectories within a room."""
    plt, ax = _setup_axes(ax, room)

    size = _room_size(room, ax)
    _draw_room(ax, size)

    src_traj = _as_trajectory(src_traj, ax)
    mic_traj = _as_trajectory(mic_traj, ax)

    _plot_trajectories(ax, src_traj, step=step, label="source path")
    _plot_trajectories(ax, mic_traj, step=step, label="mic path")

    if title:
        ax.set_title(title)
    ax.legend(loc="best")
    if show:
        plt.show()
    return ax


def _setup_axes(ax: Any | None, room: Room | Sequence[float] | Tensor) -> tuple[Any, Any]:
    """Create 2D/3D axes based on room dimension."""
    import matplotlib.pyplot as plt

    size = _room_size(room, ax)
    dim = size.numel()
    if ax is None:
        if dim == 3:
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
        else:
            _, ax = plt.subplots()
    return plt, ax


def _room_size(room: Room | Sequence[float] | Tensor, ax: Any | None) -> Tensor:
    """Normalize room size input to a 1D tensor."""
    if isinstance(room, Room):
        size = room.size
    else:
        size = room
    size = as_tensor(size)
    size = ensure_dim(size)
    return size


def _draw_room(ax: Any, size: Tensor) -> None:
    """Draw a 2D or 3D room outline."""
    dim = size.numel()
    if dim == 2:
        _draw_room_2d(ax, size)
    else:
        _draw_room_3d(ax, size)


def _draw_room_2d(ax: Any, size: Tensor) -> None:
    """Draw a 2D rectangular room."""
    import matplotlib.patches as patches

    rect = patches.Rectangle((0.0, 0.0), size[0].item(), size[1].item(),
                             fill=False, edgecolor="black")
    ax.add_patch(rect)
    ax.set_xlim(0, size[0].item())
    ax.set_ylim(0, size[1].item())
    ax.set_aspect("equal", adjustable="box")
    ax.set_xlabel("x")
    ax.set_ylabel("y")


def _draw_room_3d(ax: Any, size: Tensor) -> None:
    """Draw a 3D box representing the room."""
    x, y, z = size.tolist()
    corners = torch.tensor(
        [
            [0, 0, 0],
            [x, 0, 0],
            [x, y, 0],
            [0, y, 0],
            [0, 0, z],
            [x, 0, z],
            [x, y, z],
            [0, y, z],
        ],
        dtype=torch.float32,
    )
    edges = [
        (0, 1),
        (1, 2),
        (2, 3),
        (3, 0),
        (4, 5),
        (5, 6),
        (6, 7),
        (7, 4),
        (0, 4),
        (1, 5),
        (2, 6),
        (3, 7),
    ]
    for a, b in edges:
        ax.plot(
            [corners[a, 0], corners[b, 0]],
            [corners[a, 1], corners[b, 1]],
            [corners[a, 2], corners[b, 2]],
            color="black",
        )
    ax.set_xlim(0, x)
    ax.set_ylim(0, y)
    ax.set_zlim(0, z)
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    ax.set_zlabel("z")


def _extract_positions(entity: Source | MicrophoneArray | Tensor | Sequence, ax: Any | None) -> Tensor:
    """Extract positions from Source/MicrophoneArray or raw tensor."""
    if isinstance(entity, (Source, MicrophoneArray)):
        pos = entity.positions
    else:
        pos = entity
    pos = as_tensor(pos)
    if pos.ndim == 1:
        pos = pos.unsqueeze(0)
    return pos


def _scatter_positions(ax: Any, positions: Tensor, *, label: str, marker: str) -> None:
    """Scatter-plot positions in 2D or 3D."""
    if positions.numel() == 0:
        return
    dim = positions.shape[1]
    if dim == 2:
        ax.scatter(positions[:, 0], positions[:, 1], label=label, marker=marker)
    else:
        ax.scatter(positions[:, 0], positions[:, 1], positions[:, 2], label=label, marker=marker)


def _as_trajectory(traj: Tensor | Sequence, ax: Any | None) -> Tensor:
    """Validate and normalize a trajectory tensor."""
    traj = as_tensor(traj)
    if traj.ndim != 3:
        raise ValueError("trajectory must be of shape (T, N, dim)")
    return traj


def _plot_trajectories(ax: Any, traj: Tensor, *, step: int, label: str) -> None:
    """Plot trajectories for each entity."""
    if traj.numel() == 0:
        return
    dim = traj.shape[2]
    if dim == 2:
        for idx in range(traj.shape[1]):
            xy = traj[::step, idx]
            ax.plot(xy[:, 0], xy[:, 1], label=f"{label} {idx}")
    else:
        for idx in range(traj.shape[1]):
            xyz = traj[::step, idx]
            ax.plot(xyz[:, 0], xyz[:, 1], xyz[:, 2], label=f"{label} {idx}")



from __future__ import annotations

"""Core RIR simulation functions (static and dynamic)."""

import math
from typing import Optional, Tuple

import torch
from torch import Tensor

from .config import SimulationConfig, default_config
from .directivity import directivity_gain, split_directivity
from .room import MicrophoneArray, Room, Source
from .utils import (
    as_tensor,
    ensure_dim,
    estimate_beta_from_t60,
    estimate_t60_from_beta,
    infer_device_dtype,
    normalize_orientation,
    orientation_to_unit,
    resolve_device,
)


def simulate_rir(
    *,
    room: Room,
    sources: Source | Tensor,
    mics: MicrophoneArray | Tensor,
    max_order: int | None,
    nb_img: Optional[Tensor | Tuple[int, ...]] = None,
    nsample: Optional[int] = None,
    tmax: Optional[float] = None,
    tdiff: Optional[float] = None,
    directivity: str | tuple[str, str] | None = "omni",
    orientation: Optional[Tensor | tuple[Tensor, Tensor]] = None,
    config: Optional[SimulationConfig] = None,
    device: Optional[torch.device | str] = None,
    dtype: Optional[torch.dtype] = None,
) -> Tensor:
    """Simulate a static RIR using the image source method.

    Args:
        room: Room configuration (geometry, fs, reflection coefficients).
        sources: Source positions or a Source object.
        mics: Microphone positions or a MicrophoneArray object.
        max_order: Maximum reflection order (uses config if None).
        nb_img: Optional per-dimension image counts (overrides max_order).
        nsample: Output length in samples.
        tmax: Output length in seconds (used if nsample is None).
        tdiff: Optional time to start diffuse tail modeling.
        directivity: Directivity pattern(s) for source and mic (uses config if None).
        orientation: Orientation vectors or angles.
        config: Optional simulation configuration overrides.
        device: Output device.
        dtype: Output dtype.

    Returns:
        Tensor of shape (n_src, n_mic, nsample).
    """
    cfg = config or default_config()
    cfg.validate()

    if device is None and cfg.device is not None:
        device = cfg.device

    if max_order is None:
        if cfg.max_order is None:
            raise ValueError("max_order must be provided if not set in config")
        max_order = cfg.max_order

    if tmax is None and nsample is None and cfg.tmax is not None:
        tmax = cfg.tmax

    if directivity is None:
        directivity = cfg.directivity or "omni"

    if not isinstance(room, Room):
        raise TypeError("room must be a Room instance")
    if nsample is None and tmax is None:
        raise ValueError("nsample or tmax must be provided")
    if nsample is None:
        nsample = int(math.ceil(tmax * room.fs))
    if nsample <= 0:
        raise ValueError("nsample must be positive")
    if max_order < 0:
        raise ValueError("max_order must be non-negative")

    if isinstance(device, str):
        device = resolve_device(device)

    src_pos, src_ori = _prepare_entities(
        sources, orientation, which="source", device=device, dtype=dtype
    )
    mic_pos, mic_ori = _prepare_entities(
        mics, orientation, which="mic", device=device, dtype=dtype
    )

    device, dtype = infer_device_dtype(
        src_pos, mic_pos, room.size, device=device, dtype=dtype
    )
    src_pos = as_tensor(src_pos, device=device, dtype=dtype)
    mic_pos = as_tensor(mic_pos, device=device, dtype=dtype)

    if src_ori is not None:
        src_ori = as_tensor(src_ori, device=device, dtype=dtype)
    if mic_ori is not None:
        mic_ori = as_tensor(mic_ori, device=device, dtype=dtype)

    room_size = as_tensor(room.size, device=device, dtype=dtype)
    room_size = ensure_dim(room_size)
    dim = room_size.numel()

    if src_pos.ndim == 1:
        src_pos = src_pos.unsqueeze(0)
    if mic_pos.ndim == 1:
        mic_pos = mic_pos.unsqueeze(0)
    if src_pos.ndim != 2 or src_pos.shape[1] != dim:
        raise ValueError("sources must be of shape (n_src, dim)")
    if mic_pos.ndim != 2 or mic_pos.shape[1] != dim:
        raise ValueError("mics must be of shape (n_mic, dim)")

    beta = _resolve_beta(room, room_size, device=device, dtype=dtype)
    beta = _validate_beta(beta, dim)

    n_vec = _image_source_indices(max_order, dim, device=device, nb_img=nb_img)
    refl = _reflection_coefficients(n_vec, beta)

    src_pattern, mic_pattern = split_directivity(directivity)
    mic_dir = None
    if mic_pattern != "omni":
        if mic_ori is None:
            raise ValueError("mic orientation required for non-omni directivity")
        mic_dir = orientation_to_unit(mic_ori, dim)

    n_src = src_pos.shape[0]
    n_mic = mic_pos.shape[0]
    rir = torch.zeros((n_src, n_mic, nsample), device=device, dtype=dtype)
    fdl = cfg.frac_delay_length
    fdl2 = (fdl - 1) // 2
    img_chunk = cfg.image_chunk_size
    if img_chunk <= 0:
        img_chunk = n_vec.shape[0]

    src_dirs = None
    if src_pattern != "omni":
        if src_ori is None:
            raise ValueError("source orientation required for non-omni directivity")
        src_dirs = orientation_to_unit(src_ori, dim)
        if src_dirs.ndim == 1:
            src_dirs = src_dirs.unsqueeze(0).repeat(n_src, 1)
        if src_dirs.ndim != 2 or src_dirs.shape[0] != n_src:
            raise ValueError("source orientation must match number of sources")

    for start in range(0, n_vec.shape[0], img_chunk):
        end = min(start + img_chunk, n_vec.shape[0])
        n_vec_chunk = n_vec[start:end]
        refl_chunk = refl[start:end]
        sample_chunk, attenuation_chunk = _compute_image_contributions_batch(
            src_pos,
            mic_pos,
            room_size,
            n_vec_chunk,
            refl_chunk,
            room,
            fdl2,
            src_pattern=src_pattern,
            mic_pattern=mic_pattern,
            src_dirs=src_dirs,
            mic_dir=mic_dir,
        )
        _accumulate_rir_batch(rir, sample_chunk, attenuation_chunk, cfg)

    if tdiff is not None and tmax is not None and tdiff < tmax:
        rir = _apply_diffuse_tail(rir, room, beta, tdiff, tmax, seed=cfg.seed)
    return rir


def simulate_dynamic_rir(
    *,
    room: Room,
    src_traj: Tensor,
    mic_traj: Tensor,
    max_order: int | None,
    nsample: Optional[int] = None,
    tmax: Optional[float] = None,
    directivity: str | tuple[str, str] | None = "omni",
    orientation: Optional[Tensor | tuple[Tensor, Tensor]] = None,
    config: Optional[SimulationConfig] = None,
    device: Optional[torch.device | str] = None,
    dtype: Optional[torch.dtype] = None,
) -> Tensor:
    """Simulate time-varying RIRs for source/mic trajectories.

    Args:
        room: Room configuration.
        src_traj: Source trajectory (T, n_src, dim).
        mic_traj: Microphone trajectory (T, n_mic, dim).
        max_order: Maximum reflection order (uses config if None).
        nsample: Output length in samples.
        tmax: Output length in seconds (used if nsample is None).
        directivity: Directivity pattern(s) for source and mic (uses config if None).
        orientation: Orientation vectors or angles.
        config: Optional simulation configuration overrides.
        device: Output device.
        dtype: Output dtype.

    Returns:
        Tensor of shape (T, n_src, n_mic, nsample).
    """
    cfg = config or default_config()
    cfg.validate()

    if device is None and cfg.device is not None:
        device = cfg.device

    if max_order is None:
        if cfg.max_order is None:
            raise ValueError("max_order must be provided if not set in config")
        max_order = cfg.max_order

    if tmax is None and nsample is None and cfg.tmax is not None:
        tmax = cfg.tmax

    if directivity is None:
        directivity = cfg.directivity or "omni"

    if isinstance(device, str):
        device = resolve_device(device)

    src_traj = as_tensor(src_traj, device=device, dtype=dtype)
    mic_traj = as_tensor(mic_traj, device=device, dtype=dtype)

    if src_traj.ndim == 2:
        src_traj = src_traj.unsqueeze(1)
    if mic_traj.ndim == 2:
        mic_traj = mic_traj.unsqueeze(1)
    if src_traj.ndim != 3:
        raise ValueError("src_traj must be of shape (T, n_src, dim)")
    if mic_traj.ndim != 3:
        raise ValueError("mic_traj must be of shape (T, n_mic, dim)")
    if src_traj.shape[0] != mic_traj.shape[0]:
        raise ValueError("src_traj and mic_traj must have the same time length")

    t_steps = src_traj.shape[0]
    rirs = []
    for t_idx in range(t_steps):
        rir = simulate_rir(
            room=room,
            sources=src_traj[t_idx],
            mics=mic_traj[t_idx],
            max_order=max_order,
            nsample=nsample,
            tmax=tmax,
            directivity=directivity,
            orientation=orientation,
            config=config,
            device=device,
            dtype=dtype,
        )
        rirs.append(rir)
    return torch.stack(rirs, dim=0)


def _prepare_entities(
    entities: Source | MicrophoneArray | Tensor,
    orientation: Optional[Tensor | tuple[Tensor, Tensor]],
    *,
    which: str,
    device: Optional[torch.device | str],
    dtype: Optional[torch.dtype],
) -> Tuple[Tensor, Optional[Tensor]]:
    """Extract positions and orientations from entities or raw tensors.

    Returns:
        Tuple of (positions, orientation).
    """
    if isinstance(entities, (Source, MicrophoneArray)):
        pos = entities.positions
        ori = entities.orientation
    else:
        pos = entities
        ori = None
    if orientation is not None:
        if isinstance(orientation, (list, tuple)):
            if len(orientation) != 2:
                raise ValueError("orientation tuple must have length 2")
            ori = orientation[0] if which == "source" else orientation[1]
        else:
            ori = orientation
    pos = as_tensor(pos, device=device, dtype=dtype)
    if ori is not None:
        ori = as_tensor(ori, device=device, dtype=dtype)
    return pos, ori


def _resolve_beta(
    room: Room, room_size: Tensor, *, device: torch.device, dtype: torch.dtype
) -> Tensor:
    """Resolve reflection coefficients from beta/t60/defaults.

    Returns:
        Tensor of reflection coefficients per wall.
    """
    if room.beta is not None:
        return as_tensor(room.beta, device=device, dtype=dtype)
    if room.t60 is not None:
        return estimate_beta_from_t60(room_size, room.t60, device=device, dtype=dtype)
    dim = room_size.numel()
    default_faces = 4 if dim == 2 else 6
    return torch.ones((default_faces,), device=device, dtype=dtype)


def _validate_beta(beta: Tensor, dim: int) -> Tensor:
    """Validate beta size against room dimension.

    Returns:
        The validated beta tensor.
    """
    expected = 4 if dim == 2 else 6
    if beta.numel() != expected:
        raise ValueError(f"beta must have {expected} elements for {dim}D")
    return beta


def _image_source_indices(
    max_order: int,
    dim: int,
    *,
    device: torch.device,
    nb_img: Optional[Tensor | Tuple[int, ...]] = None,
) -> Tensor:
    """Generate image source index vectors up to the given order.

    Returns:
        Tensor of shape (n_images, dim).
    """
    if nb_img is not None:
        nb = as_tensor(nb_img, device=device, dtype=torch.int64)
        if nb.numel() != dim:
            raise ValueError("nb_img must match room dimension")
        ranges = [torch.arange(-n, n + 1, device=device, dtype=torch.int64) for n in nb]
        grids = torch.meshgrid(*ranges, indexing="ij")
        return torch.stack([g.reshape(-1) for g in grids], dim=-1)
    rng = torch.arange(-max_order, max_order + 1, device=device, dtype=torch.int64)
    grids = torch.meshgrid(*([rng] * dim), indexing="ij")
    n_vec = torch.stack([g.reshape(-1) for g in grids], dim=-1)
    order = torch.sum(torch.abs(n_vec), dim=-1)
    return n_vec[order <= max_order]


def _image_positions(src: Tensor, room_size: Tensor, n_vec: Tensor) -> Tensor:
    """Compute image source positions for a given source.

    Returns:
        Tensor of image positions (n_images, dim).
    """
    n_vec_f = n_vec.to(dtype=src.dtype)
    sign = torch.where((n_vec % 2) == 0, 1.0, -1.0).to(dtype=src.dtype)
    n = torch.floor_divide(n_vec + 1, 2).to(dtype=src.dtype)
    return 2.0 * room_size * n + sign * src


def _image_positions_batch(src_pos: Tensor, room_size: Tensor, n_vec: Tensor) -> Tensor:
    """Compute image source positions for multiple sources."""
    sign = torch.where((n_vec % 2) == 0, 1.0, -1.0).to(dtype=src_pos.dtype)
    n = torch.floor_divide(n_vec + 1, 2).to(dtype=src_pos.dtype)
    base = 2.0 * room_size * n
    return base[None, :, :] + sign[None, :, :] * src_pos[:, None, :]


def _reflection_coefficients(n_vec: Tensor, beta: Tensor) -> Tensor:
    """Compute reflection coefficients for each image source.

    Returns:
        Tensor of shape (n_images,) with per-image gains.
    """
    dim = n_vec.shape[1]
    beta = beta.view(dim, 2)
    beta_lo = beta[:, 0]
    beta_hi = beta[:, 1]

    n = n_vec
    k = torch.abs(n)
    n_hi = torch.where(n >= 0, (n + 1) // 2, k // 2)
    n_lo = torch.where(n >= 0, n // 2, (k + 1) // 2)

    n_hi = n_hi.to(dtype=beta.dtype)
    n_lo = n_lo.to(dtype=beta.dtype)

    coeff = (beta_hi**n_hi) * (beta_lo**n_lo)
    return torch.prod(coeff, dim=1)


def _compute_image_contributions(
    src: Tensor,
    mic_pos: Tensor,
    room_size: Tensor,
    n_vec: Tensor,
    refl: Tensor,
    room: Room,
    fdl2: int,
    *,
    src_pattern: str,
    mic_pattern: str,
    src_dir: Optional[Tensor],
    mic_dir: Optional[Tensor],
) -> Tuple[Tensor, Tensor]:
    """Compute sample positions and attenuation for a source and all mics."""
    img = _image_positions(src, room_size, n_vec)
    vec = mic_pos[:, None, :] - img[None, :, :]
    dist = torch.linalg.norm(vec, dim=-1)
    dist = torch.clamp(dist, min=1e-6)
    time = dist / room.c
    time = time + (fdl2 / room.fs)
    sample = time * room.fs

    gain = refl[None, :]
    if src_pattern != "omni":
        if src_dir is None:
            raise ValueError("source orientation required for non-omni directivity")
        cos_theta = _cos_between(vec, src_dir)
        gain = gain * directivity_gain(src_pattern, cos_theta)
    if mic_pattern != "omni":
        if mic_dir is None:
            raise ValueError("mic orientation required for non-omni directivity")
        cos_theta = _cos_between(-vec, mic_dir)
        gain = gain * directivity_gain(mic_pattern, cos_theta)

    attenuation = gain / dist
    return sample, attenuation


def _compute_image_contributions_batch(
    src_pos: Tensor,
    mic_pos: Tensor,
    room_size: Tensor,
    n_vec: Tensor,
    refl: Tensor,
    room: Room,
    fdl2: int,
    *,
    src_pattern: str,
    mic_pattern: str,
    src_dirs: Optional[Tensor],
    mic_dir: Optional[Tensor],
) -> Tuple[Tensor, Tensor]:
    """Compute samples/attenuation for all sources/mics/images in batch."""
    img = _image_positions_batch(src_pos, room_size, n_vec)
    vec = mic_pos[None, :, None, :] - img[:, None, :, :]
    dist = torch.linalg.norm(vec, dim=-1)
    dist = torch.clamp(dist, min=1e-6)
    time = dist / room.c
    time = time + (fdl2 / room.fs)
    sample = time * room.fs

    gain = refl.view(1, 1, -1)
    if src_pattern != "omni":
        if src_dirs is None:
            raise ValueError("source orientation required for non-omni directivity")
        src_dirs = src_dirs[:, None, None, :]
        cos_theta = _cos_between(vec, src_dirs)
        gain = gain * directivity_gain(src_pattern, cos_theta)
    if mic_pattern != "omni":
        if mic_dir is None:
            raise ValueError("mic orientation required for non-omni directivity")
        mic_dir = mic_dir[None, :, None, :] if mic_dir.ndim == 2 else mic_dir.view(1, 1, 1, -1)
        cos_theta = _cos_between(-vec, mic_dir)
        gain = gain * directivity_gain(mic_pattern, cos_theta)

    attenuation = gain / dist
    return sample, attenuation


def _select_orientation(orientation: Tensor, idx: int, count: int, dim: int) -> Tensor:
    """Pick the correct orientation vector for a given entity index."""
    if orientation.ndim == 0:
        return orientation_to_unit(orientation, dim)
    if orientation.ndim == 1:
        return orientation_to_unit(orientation, dim)
    if orientation.ndim == 2 and orientation.shape[0] == count:
        return orientation_to_unit(orientation[idx], dim)
    raise ValueError("orientation must be shape (dim,), (count, dim), or angles")


def _cos_between(vec: Tensor, orientation: Tensor) -> Tensor:
    """Compute cosine between direction vectors and orientation."""
    orientation = normalize_orientation(orientation)
    unit = vec / torch.linalg.norm(vec, dim=-1, keepdim=True)
    return torch.sum(unit * orientation, dim=-1)


def _accumulate_rir(
    rir: Tensor, sample: Tensor, amplitude: Tensor, cfg: SimulationConfig
) -> None:
    """Accumulate fractional-delay contributions into the RIR tensor."""
    idx0 = torch.floor(sample).to(torch.int64)
    frac = sample - idx0.to(sample.dtype)

    n_mic, nsample = rir.shape
    fdl = cfg.frac_delay_length
    lut_gran = cfg.sinc_lut_granularity
    use_lut = cfg.use_lut and rir.device.type != "mps"
    fdl2 = (fdl - 1) // 2

    dtype = amplitude.dtype
    n = _get_fdl_grid(fdl, device=rir.device, dtype=dtype)
    offsets = _get_fdl_offsets(fdl, device=rir.device)
    window = _get_fdl_window(fdl, device=rir.device, dtype=dtype)

    if use_lut:
        sinc_lut = _get_sinc_lut(fdl, lut_gran, device=rir.device, dtype=dtype)

    mic_offsets = (torch.arange(n_mic, device=rir.device, dtype=torch.int64) * nsample).view(
        n_mic, 1, 1
    )
    rir_flat = rir.view(-1)

    chunk_size = cfg.accumulate_chunk_size
    n_img = idx0.shape[1]
    for start in range(0, n_img, chunk_size):
        end = min(start + chunk_size, n_img)
        idx = idx0[:, start:end]
        amp = amplitude[:, start:end]
        frac_m = frac[:, start:end]

        if use_lut:
            x_off_frac = (1.0 - frac_m) * lut_gran
            lut_gran_off = torch.floor(x_off_frac).to(torch.int64)
            x_off = x_off_frac - lut_gran_off.to(dtype)
            lut_pos = lut_gran_off[..., None] + (n[None, None, :].to(torch.int64) * lut_gran)

            s0 = torch.take(sinc_lut, lut_pos)
            s1 = torch.take(sinc_lut, lut_pos + 1)
            interp = s0 + x_off[..., None] * (s1 - s0)
            filt = interp * window[None, None, :]
        else:
            t = n[None, None, :] - fdl2 - frac_m[..., None]
            filt = torch.sinc(t) * window[None, None, :]

        contrib = amp[..., None] * filt
        target = idx[..., None] + offsets[None, None, :]
        valid = (target >= 0) & (target < nsample)
        if not valid.any():
            continue

        target = target + mic_offsets
        target_flat = target[valid].to(torch.int64)
        values_flat = contrib[valid]
        rir_flat.scatter_add_(0, target_flat, values_flat)


def _accumulate_rir_batch(
    rir: Tensor, sample: Tensor, amplitude: Tensor, cfg: SimulationConfig
) -> None:
    """Accumulate fractional-delay contributions for all sources/mics."""
    fn = _get_accumulate_fn(cfg, rir.device, amplitude.dtype)
    return fn(rir, sample, amplitude)


def _accumulate_rir_batch_impl(
    rir: Tensor,
    sample: Tensor,
    amplitude: Tensor,
    *,
    fdl: int,
    lut_gran: int,
    use_lut: bool,
    chunk_size: int,
) -> None:
    """Implementation for batch accumulation (optionally compiled)."""
    idx0 = torch.floor(sample).to(torch.int64)
    frac = sample - idx0.to(sample.dtype)

    n_src, n_mic, nsample = rir.shape
    n_sm = n_src * n_mic
    idx0 = idx0.view(n_sm, -1)
    frac = frac.view(n_sm, -1)
    amplitude = amplitude.view(n_sm, -1)

    fdl2 = (fdl - 1) // 2

    n = _get_fdl_grid(fdl, device=rir.device, dtype=sample.dtype)
    offsets = _get_fdl_offsets(fdl, device=rir.device)
    window = _get_fdl_window(fdl, device=rir.device, dtype=sample.dtype)

    if use_lut:
        sinc_lut = _get_sinc_lut(fdl, lut_gran, device=rir.device, dtype=sample.dtype)

    sm_offsets = (torch.arange(n_sm, device=rir.device, dtype=torch.int64) * nsample).view(
        n_sm, 1, 1
    )
    rir_flat = rir.view(-1)

    n_img = idx0.shape[1]
    for start in range(0, n_img, chunk_size):
        end = min(start + chunk_size, n_img)
        idx = idx0[:, start:end]
        amp = amplitude[:, start:end]
        frac_m = frac[:, start:end]

        if use_lut:
            x_off_frac = (1.0 - frac_m) * lut_gran
            lut_gran_off = torch.floor(x_off_frac).to(torch.int64)
            x_off = x_off_frac - lut_gran_off.to(sample.dtype)
            lut_pos = lut_gran_off[..., None] + (n[None, None, :].to(torch.int64) * lut_gran)

            s0 = torch.take(sinc_lut, lut_pos)
            s1 = torch.take(sinc_lut, lut_pos + 1)
            interp = s0 + x_off[..., None] * (s1 - s0)
            filt = interp * window[None, None, :]
        else:
            t = n[None, None, :] - fdl2 - frac_m[..., None]
            filt = torch.sinc(t) * window[None, None, :]

        contrib = amp[..., None] * filt
        target = idx[..., None] + offsets[None, None, :]
        valid = (target >= 0) & (target < nsample)
        if not valid.any():
            continue

        target = target + sm_offsets
        target_flat = target[valid].to(torch.int64)
        values_flat = contrib[valid]
        rir_flat.scatter_add_(0, target_flat, values_flat)


_SINC_LUT_CACHE: dict[tuple[int, int, str, torch.dtype], Tensor] = {}
_FDL_GRID_CACHE: dict[tuple[int, str, torch.dtype], Tensor] = {}
_FDL_OFFSETS_CACHE: dict[tuple[int, str], Tensor] = {}
_FDL_WINDOW_CACHE: dict[tuple[int, str, torch.dtype], Tensor] = {}
_ACCUM_BATCH_COMPILED: dict[tuple[str, torch.dtype, int, int, bool, int], callable] = {}


def _get_accumulate_fn(
    cfg: SimulationConfig, device: torch.device, dtype: torch.dtype
) -> callable:
    """Return an accumulation function with config-bound constants."""
    use_lut = cfg.use_lut and device.type != "mps"
    fdl = cfg.frac_delay_length
    lut_gran = cfg.sinc_lut_granularity
    chunk_size = cfg.accumulate_chunk_size

    def _fn(rir: Tensor, sample: Tensor, amplitude: Tensor) -> None:
        _accumulate_rir_batch_impl(
            rir,
            sample,
            amplitude,
            fdl=fdl,
            lut_gran=lut_gran,
            use_lut=use_lut,
            chunk_size=chunk_size,
        )

    if device.type not in ("cuda", "mps") or not cfg.use_compile:
        return _fn
    key = (str(device), dtype, fdl, lut_gran, use_lut, chunk_size)
    compiled = _ACCUM_BATCH_COMPILED.get(key)
    if compiled is None:
        compiled = torch.compile(_fn, dynamic=True)
        _ACCUM_BATCH_COMPILED[key] = compiled
    return compiled


def _get_fdl_grid(fdl: int, *, device: torch.device, dtype: torch.dtype) -> Tensor:
    key = (fdl, str(device), dtype)
    cached = _FDL_GRID_CACHE.get(key)
    if cached is None:
        cached = torch.arange(fdl, device=device, dtype=dtype)
        _FDL_GRID_CACHE[key] = cached
    return cached


def _get_fdl_offsets(fdl: int, *, device: torch.device) -> Tensor:
    key = (fdl, str(device))
    cached = _FDL_OFFSETS_CACHE.get(key)
    if cached is None:
        fdl2 = (fdl - 1) // 2
        cached = torch.arange(fdl, device=device, dtype=torch.int64) - fdl2
        _FDL_OFFSETS_CACHE[key] = cached
    return cached


def _get_fdl_window(fdl: int, *, device: torch.device, dtype: torch.dtype) -> Tensor:
    key = (fdl, str(device), dtype)
    cached = _FDL_WINDOW_CACHE.get(key)
    if cached is None:
        cached = torch.hann_window(fdl, periodic=False, device=device, dtype=dtype)
        _FDL_WINDOW_CACHE[key] = cached
    return cached


def _get_sinc_lut(fdl: int, lut_gran: int, *, device: torch.device, dtype: torch.dtype) -> Tensor:
    """Create a sinc lookup table for fractional delays."""
    key = (fdl, lut_gran, str(device), dtype)
    cached = _SINC_LUT_CACHE.get(key)
    if cached is not None:
        return cached
    fdl2 = (fdl - 1) // 2
    lut_size = (fdl + 1) * lut_gran + 1
    n = torch.linspace(-fdl2 - 1, fdl2 + 1, lut_size, device=device, dtype=dtype)
    cached = torch.sinc(n)
    _SINC_LUT_CACHE[key] = cached
    return cached


def _apply_diffuse_tail(
    rir: Tensor,
    room: Room,
    beta: Tensor,
    tdiff: float,
    tmax: float,
    *,
    seed: Optional[int] = None,
) -> Tensor:
    """Apply a diffuse reverberation tail after tdiff.

    Returns:
        RIR tensor with diffuse tail applied.
    """
    nsample = rir.shape[-1]
    tdiff_idx = min(nsample - 1, int(math.floor(tdiff * room.fs)))
    if tdiff_idx <= 0:
        return rir
    tail_len = nsample - tdiff_idx
    t = torch.arange(tail_len, device=rir.device, dtype=rir.dtype) / room.fs

    t60 = estimate_t60_from_beta(room.size, beta)
    if math.isinf(t60) or t60 <= 0:
        decay = torch.exp(-t * 3.0)
    else:
        tau = t60 / 6.9078
        decay = torch.exp(-t / tau)

    gen = torch.Generator(device=rir.device)
    gen.manual_seed(0 if seed is None else seed)
    noise = torch.randn(rir[..., tdiff_idx:].shape, device=rir.device, dtype=rir.dtype, generator=gen)
    scale = torch.linalg.norm(rir[..., tdiff_idx - 1 : tdiff_idx], dim=-1, keepdim=True) + 1e-8
    rir[..., tdiff_idx:] = noise * decay * scale
    return rir



from __future__ import annotations

"""CMU ARCTIC dataset helpers."""

import tarfile
import urllib.request
from dataclasses import dataclass
from pathlib import Path
from typing import List, Tuple

import torch
import logging

BASE_URL = "http://www.festvox.org/cmu_arctic/packed"
VALID_SPEAKERS = {
    "aew",
    "ahw",
    "aup",
    "awb",
    "axb",
    "bdl",
    "clb",
    "eey",
    "fem",
    "gka",
    "jmk",
    "ksp",
    "ljm",
    "lnh",
    "rms",
    "rxr",
    "slp",
    "slt",
}

logger = logging.getLogger(__name__)


def list_cmu_arctic_speakers() -> List[str]:
    """Return supported CMU ARCTIC speaker IDs."""
    return sorted(VALID_SPEAKERS)


@dataclass
class CmuArcticSentence:
    """Sentence metadata from CMU ARCTIC."""
    utterance_id: str
    text: str


class CmuArcticDataset:
    def __init__(self, root: Path, speaker: str = "bdl", download: bool = False) -> None:
        """Initialize a CMU ARCTIC dataset handle.

        Args:
            root: Root directory where the dataset is stored.
            speaker: Speaker ID (e.g., "bdl").
            download: Download and extract if missing.
        """
        if speaker not in VALID_SPEAKERS:
            raise ValueError(f"unsupported speaker: {speaker}")
        self.root = Path(root)
        self.speaker = speaker
        self._base_dir = self.root / "ARCTIC"
        self._archive_name = f"cmu_us_{speaker}_arctic.tar.bz2"
        self._dataset_dir = self._base_dir / f"cmu_us_{speaker}_arctic"

        if download:
            self._download_and_extract()

        if not self._dataset_dir.exists():
            raise FileNotFoundError(
                "dataset not found; run with download=True or place the archive under "
                f"{self._base_dir}"
            )

    @property
    def wav_dir(self) -> Path:
        """Return the directory containing wav files."""
        return self._dataset_dir / "wav"

    @property
    def text_path(self) -> Path:
        """Return the path to txt.done.data."""
        return self._dataset_dir / "etc" / "txt.done.data"

    def _download_and_extract(self) -> None:
        """Download and extract the speaker archive if needed."""
        self._base_dir.mkdir(parents=True, exist_ok=True)
        archive_path = self._base_dir / self._archive_name
        url = f"{BASE_URL}/{self._archive_name}"

        if not archive_path.exists():
            logger.info("Downloading %s", url)
            _download(url, archive_path)
        if not self._dataset_dir.exists():
            logger.info("Extracting %s", archive_path)
            try:
                with tarfile.open(archive_path, "r:bz2") as tar:
                    tar.extractall(self._base_dir)
            except (tarfile.ReadError, EOFError, OSError) as exc:
                logger.warning("Extraction failed (%s); re-downloading.", exc)
                if archive_path.exists():
                    archive_path.unlink()
                _download(url, archive_path)
                with tarfile.open(archive_path, "r:bz2") as tar:
                    tar.extractall(self._base_dir)

    def sentences(self) -> List[CmuArcticSentence]:
        """Parse all sentence metadata."""
        sentences: List[CmuArcticSentence] = []
        with self.text_path.open("r", encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                utt, text = _parse_text_line(line)
                sentences.append(CmuArcticSentence(utterance_id=utt, text=text))
        return sentences

    def available_sentences(self) -> List[CmuArcticSentence]:
        """Return sentences that have a corresponding wav file."""
        wav_ids = {p.stem for p in self.wav_dir.glob("*.wav")}
        return [s for s in self.sentences() if s.utterance_id in wav_ids]

    def list_speakers(self) -> List[str]:
        """Return available speaker IDs."""
        return list_cmu_arctic_speakers()

    def wav_path(self, utterance_id: str) -> Path:
        """Return the wav path for an utterance ID."""
        return self.wav_dir / f"{utterance_id}.wav"

    def load_wav(self, utterance_id: str) -> Tuple[torch.Tensor, int]:
        """Load a mono wav for the given utterance ID."""
        path = self.wav_path(utterance_id)
        return load_wav_mono(path)


def _download(url: str, dest: Path, retries: int = 1) -> None:
    """Download a file with retry and resume-safe temp file."""
    for attempt in range(retries + 1):
        try:
            _stream_download(url, dest)
            return
        except Exception as exc:
            if dest.exists():
                dest.unlink()
            if attempt >= retries:
                raise
            logger.warning("Download failed (%s); retrying...", exc)


def _stream_download(url: str, dest: Path) -> None:
    """Stream a URL to disk with a progress indicator."""
    tmp_path = dest.with_suffix(dest.suffix + ".part")
    if tmp_path.exists():
        tmp_path.unlink()

    with urllib.request.urlopen(url) as response:
        total = response.length or 0
        downloaded = 0
        chunk_size = 1024 * 1024
        with tmp_path.open("wb") as f:
            while True:
                chunk = response.read(chunk_size)
                if not chunk:
                    break
                f.write(chunk)
                downloaded += len(chunk)
    if total > 0 and downloaded != total:
        raise IOError(f"incomplete download: {downloaded} of {total} bytes")
    tmp_path.replace(dest)


def _parse_text_line(line: str) -> Tuple[str, str]:
    """Parse a txt.done.data line into (utterance_id, text)."""
    left, _, right = line.partition('"')
    utterance = left.replace("(", "").strip().split()[0]
    text = right.rsplit('"', 1)[0]
    return utterance, text


def load_wav_mono(path: Path) -> Tuple[torch.Tensor, int]:
    """Load a wav file and return mono audio and sample rate."""
    import soundfile as sf

    audio, sample_rate = sf.read(str(path), dtype="float32", always_2d=True)
    audio_t = torch.from_numpy(audio)
    if audio_t.shape[1] > 1:
        audio_t = audio_t.mean(dim=1)
    else:
        audio_t = audio_t.squeeze(1)
    return audio_t, sample_rate


def save_wav(path: Path, audio: torch.Tensor, sample_rate: int) -> None:
    """Save a mono or multi-channel wav to disk."""
    import soundfile as sf

    audio = audio.detach().cpu().clamp(-1.0, 1.0).to(torch.float32)
    if audio.ndim == 2 and audio.shape[0] <= 8:
        audio = audio.transpose(0, 1)
    sf.write(str(path), audio.numpy(), sample_rate)



from __future__ import annotations

"""Scene container for simulation inputs."""

from dataclasses import dataclass
from typing import Optional

import torch
from torch import Tensor

from .room import MicrophoneArray, Room, Source


@dataclass(frozen=True)
class Scene:
    """Container for room, sources, microphones, and optional trajectories."""

    room: Room
    sources: Source
    mics: MicrophoneArray
    src_traj: Optional[Tensor] = None
    mic_traj: Optional[Tensor] = None

    def is_dynamic(self) -> bool:
        """Return True if any trajectory is provided."""
        return self.src_traj is not None or self.mic_traj is not None

    def validate(self) -> None:
        """Validate scene consistency and trajectory shapes."""
        if not isinstance(self.room, Room):
            raise TypeError("room must be a Room instance")
        if not isinstance(self.sources, Source):
            raise TypeError("sources must be a Source instance")
        if not isinstance(self.mics, MicrophoneArray):
            raise TypeError("mics must be a MicrophoneArray instance")

        dim = self.room.size.numel()
        n_src = self.sources.positions.shape[0]
        n_mic = self.mics.positions.shape[0]

        t_src = _validate_traj(self.src_traj, n_src, dim, "src_traj")
        t_mic = _validate_traj(self.mic_traj, n_mic, dim, "mic_traj")
        if t_src is not None and t_mic is not None and t_src != t_mic:
            raise ValueError("src_traj and mic_traj must have matching time steps")


def _validate_traj(
    traj: Optional[Tensor],
    count: int,
    dim: int,
    name: str,
) -> Optional[int]:
    if traj is None:
        return None
    if not torch.is_tensor(traj):
        raise TypeError(f"{name} must be a Tensor")
    if traj.ndim == 2:
        if count != 1:
            raise ValueError(f"{name} must have shape (T, {count}, {dim})")
        if traj.shape[1] != dim:
            raise ValueError(f"{name} must have shape (T, {dim}) for single entity")
        return traj.shape[0]
    if traj.ndim == 3:
        if traj.shape[1] != count or traj.shape[2] != dim:
            raise ValueError(f"{name} must have shape (T, {count}, {dim})")
        return traj.shape[0]
    raise ValueError(f"{name} must have shape (T, {count}, {dim})")



from __future__ import annotations

"""Dataset template for future extensions.

Work in progress:
    This module is a placeholder for future dataset integrations. The goal is
    to provide a consistent interface for downloading, caching, enumerating
    speakers/utterances, and loading audio in a reproducible way.
"""

from dataclasses import dataclass
from pathlib import Path
from typing import List, Sequence, Tuple

import torch

from .base import BaseDataset, SentenceLike


@dataclass
class TemplateSentence:
    """Minimal sentence metadata for a template dataset."""

    utterance_id: str
    text: str


class TemplateDataset(BaseDataset):
    """Work in progress template dataset implementation.

    Goal:
        Implement concrete dataset handlers by filling in download logic,
        metadata parsing, and audio loading while keeping the BaseDataset
        protocol intact.
    """

    def __init__(self, root: Path, speaker: str = "default", download: bool = False) -> None:
        self.root = Path(root)
        self.speaker = speaker
        if download:
            raise NotImplementedError(
                "download is not implemented yet. Intended to fetch and cache "
                "dataset archives under root."
            )

    def list_speakers(self) -> List[str]:
        """Return available speaker IDs."""
        return ["default"]

    def available_sentences(self) -> Sequence[SentenceLike]:
        """Return sentence entries that have audio available.

        Work in progress:
            Intended to parse dataset metadata and filter to utterances that
            have corresponding audio files on disk.
        """
        raise NotImplementedError("available_sentences is not implemented yet")

    def load_wav(self, utterance_id: str) -> Tuple[torch.Tensor, int]:
        """Load audio for an utterance and return (audio, sample_rate).

        Work in progress:
            Intended to load audio from local cache and return mono float32.
        """
        raise NotImplementedError("load_wav is not implemented yet")



from __future__ import annotations

"""Higher-level plotting utilities used by examples."""

from pathlib import Path
from typing import Any, Optional, Sequence

import torch

from .plotting import plot_scene_dynamic, plot_scene_static


def plot_scene_and_save(
    *,
    out_dir: Path,
    room: Sequence[float] | torch.Tensor,
    sources: object | torch.Tensor | Sequence,
    mics: object | torch.Tensor | Sequence,
    src_traj: Optional[torch.Tensor | Sequence] = None,
    mic_traj: Optional[torch.Tensor | Sequence] = None,
    prefix: str = "scene",
    step: int = 1,
    show: bool = False,
    plot_2d: bool = True,
    plot_3d: bool = True,
) -> tuple[list[Path], list[Path]]:
    """Plot static and dynamic scenes and save images to disk.

    Args:
        out_dir: Output directory for PNGs.
        room: Room size tensor or sequence.
        sources: Source positions or Source-like object.
        mics: Microphone positions or MicrophoneArray-like object.
        src_traj: Optional source trajectory (T, n_src, dim).
        mic_traj: Optional mic trajectory (T, n_mic, dim).
        prefix: Filename prefix for saved images.
        step: Subsampling step for trajectories.
        show: Whether to show figures interactively.
        plot_2d: Save 2D projections.
        plot_3d: Save 3D projections (only if dim == 3).

    Returns:
        Tuple of (static_paths, dynamic_paths).
    """
    out_dir = Path(out_dir)
    out_dir.mkdir(parents=True, exist_ok=True)

    room_size = _to_cpu(room)
    src_pos = _positions_to_cpu(sources)
    mic_pos = _positions_to_cpu(mics)
    dim = int(room_size.numel())

    static_paths: list[Path] = []
    dynamic_paths: list[Path] = []

    for view_dim, enabled in ((2, plot_2d), (3, plot_3d)):
        if not enabled:
            continue
        if view_dim == 2 and dim < 2:
            continue
        if view_dim == 3 and dim < 3:
            continue
        view_room = room_size[:view_dim]
        view_src = src_pos[:, :view_dim]
        view_mic = mic_pos[:, :view_dim]

        ax = plot_scene_static(
            room=view_room,
            sources=view_src,
            mics=view_mic,
            title=f"Room scene ({view_dim}D static)",
            show=False,
        )
        static_path = out_dir / f"{prefix}_static_{view_dim}d.png"
        _save_axes(ax, static_path, show=show)
        static_paths.append(static_path)

        if src_traj is not None or mic_traj is not None:
            steps = _traj_steps(src_traj, mic_traj)
            src_traj = _trajectory_to_cpu(src_traj, src_pos, steps)
            mic_traj = _trajectory_to_cpu(mic_traj, mic_pos, steps)
            view_src_traj = src_traj[:, :, :view_dim]
            view_mic_traj = mic_traj[:, :, :view_dim]
            ax = plot_scene_dynamic(
                room=view_room,
                src_traj=view_src_traj,
                mic_traj=view_mic_traj,
                step=step,
                title=f"Room scene ({view_dim}D trajectories)",
                show=False,
            )
            _overlay_positions(ax, view_src, view_mic)
            dynamic_path = out_dir / f"{prefix}_dynamic_{view_dim}d.png"
            _save_axes(ax, dynamic_path, show=show)
            dynamic_paths.append(dynamic_path)

    return static_paths, dynamic_paths


def _to_cpu(value: Any) -> torch.Tensor:
    """Move a value to CPU as a tensor."""
    if torch.is_tensor(value):
        return value.detach().cpu()
    return torch.as_tensor(value).detach().cpu()


def _positions_to_cpu(entity: torch.Tensor | object) -> torch.Tensor:
    """Extract positions from an entity and move to CPU."""
    pos = getattr(entity, "positions", entity)
    pos = _to_cpu(pos)
    if pos.ndim == 1:
        pos = pos.unsqueeze(0)
    return pos


def _traj_steps(src_traj: Optional[torch.Tensor | Sequence], mic_traj: Optional[torch.Tensor | Sequence]) -> int:
    """Infer the number of trajectory steps."""
    if src_traj is not None:
        return int(_to_cpu(src_traj).shape[0])
    return int(_to_cpu(mic_traj).shape[0])


def _trajectory_to_cpu(
    traj: Optional[torch.Tensor | Sequence], fallback_pos: torch.Tensor, steps: int
) -> torch.Tensor:
    """Normalize trajectory to CPU tensor with shape (T, N, dim)."""
    if traj is None:
        return fallback_pos.unsqueeze(0).repeat(steps, 1, 1)
    traj = _to_cpu(traj)
    if traj.ndim != 3:
        raise ValueError("trajectory must be of shape (T, N, dim)")
    return traj


def _save_axes(ax: Any, path: Path, *, show: bool) -> None:
    """Save a matplotlib axis to disk."""
    import matplotlib.pyplot as plt

    fig = ax.figure
    fig.tight_layout()
    fig.savefig(path, dpi=150)
    if show:
        plt.show()
    plt.close(fig)


def _overlay_positions(ax: Any, sources: torch.Tensor, mics: torch.Tensor) -> None:
    """Overlay static source and mic positions on an axis."""
    if sources.numel() > 0:
        if sources.shape[1] == 2:
            ax.scatter(sources[:, 0], sources[:, 1], marker="^", label="sources", color="tab:green")
        else:
            ax.scatter(
                sources[:, 0],
                sources[:, 1],
                sources[:, 2],
                marker="^",
                label="sources",
                color="tab:green",
            )
    if mics.numel() > 0:
        if mics.shape[1] == 2:
            ax.scatter(mics[:, 0], mics[:, 1], marker="o", label="mics", color="tab:orange")
        else:
            ax.scatter(
                mics[:, 0],
                mics[:, 1],
                mics[:, 2],
                marker="o",
                label="mics",
                color="tab:orange",
            )
    ax.legend(loc="best")
